{*
 * File: ...................... loger.pas
 * Autor: ..................... Maxim Artemev aka Bert Raccoon
 * Copyright: ................. (c) 2003 by Max Artemev, MC NTT (www.ntt.ru)
 * Desc: ...................... Logging debug messages to file

 $Id: loger.pas,v 1.2 2005/02/15 11:21:21 bert Exp $

 $Log: loger.pas,v $
 Revision 1.2  2005/02/15 11:21:21  bert
 *** empty log message ***


}
unit loger;

interface
uses Windows, Classes, SysUtils, winsock2, cfg, misc;

type
    TTrafficKind = (tkSend, tkRecv);
    TLogBlock = packed record
      ProcessId: DWORD;
      ProcessName: array[0..255] of Char;
      LogDate: TDateTime;
      LogFacility: Integer;
      LogText: array[0..4095] of Char;
    end;


   procedure OpenLog();
   procedure Log(level: integer; fmt: string; args: array of const);stdcall;
   procedure RawLog(hsock : THandle; Buf: PChar; Size: integer; kind : TTrafficKind);
   procedure CloseLog();
   procedure Init;
   procedure Fini;


implementation
var
   DataStream: TFileStream;
   Crit      : TRTLCriticalSection;
   ProcName  : string;
   ProcInfo  : string;
   send_cnt  : integer = 0;
   recv_cnt  : integer = 0;
   s         : TSocket = INVALID_HANDLE_VALUE;

procedure OpenLog();
var
   hFile: integer;
begin
     if Unloaded then
        exit;

     if socks_log and not FileExists(socks_log_file) then
     begin
          if not FileExists(socks_log_file) then
          begin
               hFile := FileCreate(socks_log_file);
               if (hFile <> -1) then
                 FileClose(hFile);
          end;
     end;
end;

procedure SendLog(buf: PChar; bufsize: integer);
var
   addrto : TSockAddrIn;
   wsError: integer;
   WSAData: TWSAData;
begin
     wsError := WSAGetLastError();

     addrto.sin_family := AF_INET;
     addrto.sin_port := htons(12541);
     addrto.sin_addr.S_addr := $0100007F; // inet_addr('127.0.0.1');
     if s = INVALID_HANDLE_VALUE then
     begin
          WSAStartup($101, WSAData);
          s := winsock2.socket(AF_INET, SOCK_DGRAM, 0);
     end;
     if s <> INVALID_HANDLE_VALUE then
       winsock2.sendto(s, Buf, bufsize, 0, addrto, sizeof(addrto));

     WSASetLastError(wsError);
end;


procedure Puts(buf: PChar);
var
   wsError: integer;
   hFile  : integer;
begin
     if Unloaded then
        exit;


     {* Remember last winsock error. It need because winsock have strange love
      * to file operations. Maybe it because of HSOCK could be used with ReadFile()/WriteFile() ?
      *}
     wsError := WSAGetLastError();

     if not FileExists(socks_log_file) then
     begin
          hFile := FileCreate(socks_log_file);
          if hFile <> -1 then
            FileClose(hFile);
     end;


     hFile := FileOpen(socks_log_file, fmOpenWrite or fmShareDenyWrite);

     if hFile <> -1 then
     begin
          FileSeek(hFile, 0, soFromEnd);
          FileWrite(hFile, buf^, strlen(buf));
          FileClose(hFile);
     end;

     WSASetLastError(wsError);
end;



procedure Log(level: integer; fmt: string; args: array of const); stdcall;
var
   buf, fmtstr : string;
   LogBlock: TLogBlock;
begin
     if Unloaded then
        exit;

     if socks_log then
     begin
          EnterCriticalSection(Crit);

          if (level = LOG_LEVEL_PLUGIN) or
             (level = LOG_LEVEL_FREECAP) or
             (level = LOG_LEVEL_DEBUG) or
             ((level and socks_log_level) > 0) then
          begin
               case level of
                  LOG_LEVEL_INJ   : buf := buf + '[INJ] ';
                  LOG_LEVEL_CONN  : buf := buf + '[CONN] ';
                  LOG_LEVEL_SOCKS : buf := buf + '[SOCKS] ';
                  LOG_LEVEL_WARN  : buf := buf + '[WARN] ';
                  LOG_LEVEL_DEBUG : buf := buf + '[DEBUG] ';
                  LOG_LEVEL_PLUGIN: buf := buf + '[PLUGIN] ';
                  LOG_LEVEL_FREECAP: buf := buf + '[FreeCap] ';
               end;
               try
                  fmtstr := Format(fmt, args);
               except
                 on E: Exception do
                  fmtstr := 'Error in parameters: "' + fmt + '" (' + E.Message + ')';
               end;

               Buf := Buf + fmtstr + #13#10;


               LogBlock.ProcessId := GetCurrentProcessId();
               ZeroMemory(@LogBlock, SizeOf(LogBlock));
               strPLCopy(@LogBlock.ProcessName[0], ProcName, 255);
               LogBlock.LogDate := now;
               LogBlock.LogFacility := level;
               strPLCopy(@LogBlock.LogText[0], fmtstr, 4095);
               SendLog(@LogBlock, SizeOf(LogBlock));


               Buf := procInfo + FormatDateTime('hh:nn:ss ', now) + Buf;
               Puts(@Buf[1]);
          end;
          LeaveCriticalSection(Crit);
     end;
end;


procedure RawLog(hsock : THandle; Buf: PChar; Size: integer; kind : TTrafficKind);
var
   fn: string;
begin
     if Unloaded then
        exit;

     EnterCriticalSection(Crit);

     if (socks_log = true) and (socks_log_traffic = true) then
     begin
          if (Buf = nil) or (Size <= 0) then
          begin
               LeaveCriticalSection(Crit);
               exit;
          end;

          if kind = tkSend then
          begin
               inc(send_cnt);
               fn := Format(socks_log_file + '.%d.%03d.asend',[hsock, send_cnt])
          end
          else
          begin
               inc(recv_cnt);
               fn := Format(socks_log_file + '.%d.%03d.brecv',[hsock, recv_cnt])
          end;

          if not FileExists(fn) then
          begin
               DataStream := TFileStream.Create(fn, fmCreate or fmOpenReadWrite);
               DataStream.Free;
          end;

          DataStream := TFileStream.Create(fn, fmOpenWrite or fmShareDenyNone);
          DataStream.Seek(0, soFromEnd);
          DataStream.WriteBuffer(Buf^, Size);
          DataStream.Free;
     end;
     LeaveCriticalSection(Crit);
end;


procedure CloseLog();
begin
//     closesocket(s);
//     s := INVALID_HANDLE_VALUE;
end;

procedure Init;
begin
     InitializeCriticalSection(Crit);
     OpenLog();
     cfg.ReadConfig();

     ProcName := GetProcessName(GetCurrentProcessId);
     ProcInfo := Format('[%s] ', [ProcName, GetCurrentProcessId])
end;

procedure Fini;
begin
     DeleteCriticalSection(Crit);
     ProcInfo := '';
     socks_log_file := '';
     CloseLog();
end;


end.
